function [dist,p] = dtw_(x, y, w)
%DTW dynamic time warping with Euclidean local cost for multidimensional time series
% Tries to use built-in dtw function. If you override dtw, your overriding 
% dtw function will be used. It must return the same parameters as the
% the built-in dtw function does: [dist,ix,iy].
% If dtw call fails, the implementation below is used. Note that it treats
% the warping window differently than the builtin function.
%
% Input
% x:     [N x d]               d dimensional time series of length N
% y:     [M x d]               d dimensional time series of length M
% w:     [1 x 1]   (optional)  warping window parameter. Default: w=max(N,M)
%
% Output
% dist:  [1 x 1]               dtw(x,y) with local Euclidean distance
% p:     [L x 2]   (optional)  warping path of length L
%
%
% Author: David Schultz, DAI-Lab, TU Berlin, Germany, 06-2020
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

[N,d] = size(x);
[M,d2] = size(y);

if d ~= d2
    error('Dimensions of both time series x,y have to agree.')
end

if nargin < 3 || isempty(w)
    w = max(N,M);
end


% try to use builtin
try
    [dist,ix,iy] = dtw(x',y',w,'squared');
    dist = sqrt(dist);
    p = [ix, iy];
    return
catch
end


% correct warping window if it is too small
w = max(w,abs(N-M));

C = inf(N,M);

if d==1 % runs faster than pdist2 for d==1
    C(1,1) = (x(1)-y(1))^2;
    for n = 2: min(N,w+1)
        C(n,1) = C(n-1,1) + (x(n)-y(1))^2;
    end
    for m = 2 : min(M,w+1)
        C(1,m) = C(1,m-1) + (x(1)-y(m))^2;
    end
    for n = 2:N
        for m = max(2, n-w) : min(M, n+w)
            C(n,m) = (x(n)-y(m))^2 + min( min(C(n-1,m-1), C(n-1,m) ) , C(n,m-1) );
        end
    end
else
    D = pdist2(x,y).^2;
    C(1:min(N,w+1),1) = cumsum(D(1:min(N,w+1),1));
    C(1,1:min(M,w+1)) = cumsum(D(1,1:min(M,w+1)));
    for n = 2:N
        for m = max(2, n-w) : min(M, n+w)
            C(n,m) = D(n,m) + min( min(C(n-1,m-1), C(n-1,m)), C(n,m-1));
        end
    end 
end

dist = sqrt(C(N,M));


% compute warping path p
if nargout >= 2
    n = N;
    m = M;
    p = zeros(n+m-1,2);
    p(end,:) = [n,m];
    k = 1;
    while n+m ~= 2
        if n == 1
            m = m-1;
        elseif m == 1
            n = n-1;
        else
            C_diag = C(n-1,m-1);
            C_r = C(n,m-1);
            C_d = C(n-1,m);
            if C_diag <= C_r
                if C_diag <= C_d
                    n=n-1;
                    m=m-1;
                else
                    n=n-1;
                end
            elseif C_r <= C_d
                m=m-1;
            else
                n=n-1;
            end
        end
        p(end-k,:)=[n,m]; 
        k=k+1;
    end
    p = p(end-k+1:end,:);
end

end
