function [y,dist,p] = impute_dtw_wp(x,y,v,w)
%IMPUTE_DTW_WP dynamic time warping based data imputation method
% x is a reference time series without missing values
% y is a time series with missing values encoded by NaN values or the vector v
% v indicates which values in y are missing. If y is multivariate and
% y(m,k) is NaN for some k (that is v(m)=1), the complete data point y(m,:) 
% is estimated.
%
% Input
% x:  [N x d]                d dimensional time series of length N
% y:  [M x d]                d dimensional time series of length M
% v:  [M x 1] binary  (opt)  encodes values in y to be replaced 
%                            (1=replace, 0=keep)
% w:  [1 x 1]    (optional)  warping window parameter. Default: w=max(N,M)
%
% Output
% dist:  [1 x 1]               dtw(x,y) with local Euclidean distance
% y:     [M x d]               y with inserted values
% p:     [L x 2]  (optional)   warping path of length L
%
%
% Author: David Schultz, DAI-Lab, TU Berlin, Germany, 11-2021
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


[N,d] = size(x);
[M,d2] = size(y);

if d ~= d2
    error('Dimensions of both time series x,y have to agree. dim(x)=%i, dim(y)=%i',d,d2)
end

% 
if nargin >= 3 && ~isempty(v)
    if length(v)~=M
        error('lengths of y and v have to agree. length(v)=%i. size(y)=(%i,%i)',length(v),M,d2)
    end
    y(v==1,:) = NaN;
end
if nargin < 4 || isempty(w)
    w = max(N,M);
end

v = sum(isnan(y),2)>0; % indicates for each m if y(m,:) has a NaN value 
    
% correct warping window if it is too small
w = max(w,abs(N-M));


C = inf(N,M);



if d==1 % runs faster than pdist2 for d==1
    C(1,1) = (x(1)-y(1))^2;
    for n = 2: min(N,w+1)
        C(n,1) = C(n-1,1) + (x(n)-y(1))^2;
    end
    for m = 2 : min(M,w+1)
        C(1,m) = C(1,m-1) + (x(1)-y(m))^2;
    end
    for n = 2:N
        for m = max(2, n-w) : min(M, n+w)
            local_cost = 0; % heuristically, assume zero cost if there is a NaN value
            if ~v(m) % y(m) is not NaN
                local_cost = (x(n)-y(m))^2;
            end
            C(n,m) = local_cost + min( min(C(n-1,m-1), C(n-1,m) ) , C(n,m-1) );
        end
    end
else % d > 1 (multivariate time series)
    D = pdist2(x,y).^2;
    D(isnan(D))=0; % replace NaNs by zeros (heuristically, assume zero cost)
    
    C(1:min(N,w+1),1) = cumsum(D(1:min(N,w+1),1));
    C(1,1:min(M,w+1)) = cumsum(D(1,1:min(M,w+1)));
    for n = 2:N
        for m = max(2, n-w) : min(M, n+w)
            C(n,m) = D(n,m) + min( min(C(n-1,m-1), C(n-1,m)), C(n,m-1));
        end
    end 
end



dist = sqrt(C(N,M));

% insert last element if it is NaN
if isnan(y(M))
    y(M,:)=x(N,:);
end

% compute warping path p and insert missing values in y
if nargout >= 0
    n = N;
    m = M;
    p = zeros(n+m-1,2);
    p(end,:) = [n,m];
    k = 1;
    num_up = 0;
    while n+m ~= 2
        if n == 1
            m = m-1;
            num_up = 0;
        elseif m == 1
            n = n-1;
            num_up = num_up+1;
        else
            C_diag = C(n-1,m-1);
            C_r = C(n,m-1);
            C_d = C(n-1,m);
            if C_diag <= C_r
                if C_diag <= C_d
                    n=n-1;
                    m=m-1;
                    num_up=0;
                else
                    n=n-1;
                    num_up = num_up+1;
                end
            elseif C_r <= C_d
                m=m-1;
                num_up = 0;
            else % go vertical. if nan, count how often. then set mean of x(i)'s
                n=n-1;
                num_up = num_up+1;
            end
        end
        p(end-k,:)=[n,m]; 
        k=k+1;
        if v(m)   % y(m,:) has a NaN value. Insert this missing value
            y(m,:) = mean(x(n:n+num_up,:));
        end
    end
    p = p(end-k+1:end,:);
end


end