function [idx,M,X] = kmeans_dtw_imputation(X, k, means_init, maxIter, w, mean_algo)
%KMEANS_DTW k-means algorithm wrt the Dynamic Time Warping distance for
%time series with missing values. The missing values are estimated as part
%of the algorithm.
% Input
%   X               Cell array or matrix of time series 
%                   * Cell array form: X consists of N time series x_i of size 
%                     [n_i,d], where n_i is the length of time series x_i 
%                     and d is the uniform dimension of all time series.
%                     example:   X{1} = x_1;  % size [n_1,d]
%                                X{2} = x_2;  % size [n_2,d]
%                                ...
%                                X{N} = x_N;  % size [n_N,d]
%                   * Matrix form: X is a [N x n x d] tensor, where 
%                       - N is the number of time series,
%                       - n is the length of the time series
%                       - d is the dimension of the time series
%                     - Only supports equal-length time series.
%                     - For univariate time sries (that is d==1), X is 
%                       simply a [N x n] matrix.
%
%                   * X may have missing values encoded by NaN values
%                     These missing values will be estimated
%
%   k               number of clusters
%   maxIter (opt)   maximum number of iterations. Default = 10
%   
%
%   w       (opt)   Warping window parameter for dtw distance.
%                   Default = None
%   
%
% Output
%   idx            cluster indices for each observation
%   M              cell array of means, i.e. M{i} is the mean of cluster i
%   X              cell array of sample time series with imputed values
%                  
%
%
% (o) = optional
%
% Author: David Schultz, DAI-Lab, TU Berlin, Germany, 10-2021 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% convert X to cell array if necessary
if ~iscell(X)
   if length(size(X)) > 3
       error('Invalid shape of X. size(X) must be [N,n] or [N,n,d].')
   end
   [N,n,d] = size(X);
   Y = cell(N,1);
   for i = 1:N
       Y{i} = reshape(X(i,:,:),n,d);
   end
   X=Y;
end

if nargin < 4 || isempty(maxIter)
    maxIter = 10;
end

if nargin < 5
    w = [];
end

if nargin < 6 || isempty(mean_algo)
    mean_algo = @(X)SSG(X, [], [], [], w);
end




% V = binary cell array encoding missing values. 
% V{i}(j) = 1 <=> (x_i)_j is NaN, i.e. missing
V = cell(N,1);
for i = 1:length(X)
   x = X{i};
   v = sum(isnan(x),2)>=1;
   V{i} = v;
end

X_orig = cell(N,1);

% remove missing values in X
for i = 1:length(X)
   x = X{i};
   X_orig{i} = x;
   x(isnan(x)) = [];
   X{i} = x;
end



% TODO: what if a time series is empty? -> set it to the mean of a random or largest cluster
m = length(X);
if nargin < 3 || isempty(means_init)
    P = randperm(m);
    means_init = P(1:k);
end
means = X(means_init);

idx_old = 0;
idx = 1;


counter = 0;
while ~isequal(idx_old,idx)  && counter < maxIter
    counter=counter+1;
    idx_old = idx;
    
    % update cluster assignments idx
    if k==1
        idx=ones(1,m);
    else
        D = zeros(k,m); % all distances to the centroids
        for j = 1:k
            for i=1:m
                D(j,i) = dtw_(X{i}, means{j}, w)^2;
            end
        end
        [~,idx] = min(D);
    end

    % update means and missing values
    for j = 1:k
        Xj = X(idx==j,:);
        means{j} = mean_algo(Xj);
        
        % update missing values in cluster j, store them to X
        indices_of_cluster_j = find(idx==j);

        for l = 1:length(indices_of_cluster_j)
            ind = indices_of_cluster_j(l);
            x = means{j};
            if counter==1
                y = X_orig{ind};
            else
                y = X{ind}; 
            end
            v = V{ind};
            if ismember(1,v) % y has missing values
                y = impute_dtw_wp(x,y,v,w);
            end
            X{ind} = y; 
        end
        
    end

    M=means;
    
end
