function [idx,M,cost] = kmeans_dtw(X, k, maxIter, mean_algo, w)
%KMEANS_DTW Summary of this function goes here
% Input
%   X               Cell array or matrix of time series
%                   * Cell array form: X consists of N time series x_i of size 
%                     [n_i,d], where n_i is the length of time series x_i 
%                     and d is the uniform dimension of all time series.
%                     example:   X{1} = x_1;  % size [n_1,d]
%                                X{2} = x_2;  % size [n_2,d]
%                                ...
%                                X{N} = x_N;  % size [n_N,d]
%                   * Matrix form: X is a [N x n x d] tensor, where 
%                       - N is the number of time series,
%                       - n is the length of the time series
%                       - d is the dimension of the time series
%                     - Only supports equal-length time series.
%                     - For univariate time sries (that is d==1), X is 
%                       simply a [N x n] matrix.
%
%   k               number of clusters
%   maxIter         maximum number of iterations
%   (o)
%
%   w              Warping window parameter for dtw distance.
%   (o)
%
% Output
%   idx            cluster indices for each observation
%   M              struct of means, i.e. M{i} is the mean of cluster i
%   cost           sum of squared dtw distances of datapoints to their means
%
%
% (o) = optional
%
% Author: David Schultz, DAI-Lab, TU Berlin, Germany, 06-2020 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


% convert X to cell array if necessary
if ~iscell(X)
   if length(size(X)) > 3
       error('Invalid shape of X. size(X) must be [N,n] or [N,n,d].')
   end
   [N,n,d] = size(X);
   Y = cell(N);
   for i = 1:N
       Y{i} = reshape(X(i,:,:),n,d);
   end
   X=Y;
end

if nargin < 3 || isempty(maxIter)
    maxIter = 100;
end

if nargin < 4 || isempty(mean_algo)
    mean_algo = @MM;
end

if nargin < 5
    w = [];
end

m = length(X);
P = randperm(m);
means = X(P(1:k));

idx_old = 0;
idx = 1;

counter = 0;
while ~isequal(idx_old,idx) && counter < maxIter
    counter=counter+1;
    idx_old = idx;
    
    % update assignments idx
    D = zeros(k,m); % all distances to the centroids
    for j = 1:k
        for i=1:m
            D(j,i) = dtw_(X{i}, means{j},w)^2;
        end
    end
    [~,idx] = min(D);

    % update means
    for j = 1:k
        Xj = X(idx==j);
        if ~ isempty(Xj)
            means{j} = mean_algo(Xj);
        end
    end

    M=means;
    
    if nargout > 2
        cost = 0;
        for i = 1:m
            j =  idx(i);
            cost = cost + dtw_(X{i},means{j},w)^2;
        end
    end
end
