function [mu,f] = MM(X, maxIter, tol, initSequence, window)
%MM Majorize Minimize algorithm for the Fechet sample mean problem under
% the Euclidean dynamic time warping distance.
% 
% MM aims at finding a sample mean under dynamic time warping of  
% sample X = (x_1,...,x_N) containing time series x_i, each of arbitraty 
% length and uniform dimension d. Similarity between the
% d-dimensional data points x_ij is measured by means of the Euclidean 
% distance. Other local distances are not supported.
%
% Mathematically this can be formulated as an optimization problem.
% A sample mean mu is a time series that satisfies 
%
%    F(mu) = min F(x),
%            x
%
% where F is the Frechet function, defined by
%
%            N
%    F(x) = sum dtw^2(x,x_i).
%           i=1
%
% 
% The algorithm requires to choose an initial (mean candidate) sequence. 
% Then MM iteratively refines it with respect to a majorize-minimize
% optimization scheme. Details are discussed in [1]. The MM algorithm is
% equivalent to the DBA algorithm [2].
% 
%
%
% Input
%   X               Cell array or matrix of time series
%                   * Cell array form: X consists of N time series x_i of size 
%                     [n_i,d], where n_i is the length of time series x_i 
%                     and d is the uniform dimension of all time series.
%                     example:   X{1} = x_1;  % size [n_1,d]
%                                X{2} = x_2;  % size [n_2,d]
%                                ...
%                                X{N} = x_N;  % size [n_N,d]
%                   * Matrix form: X is a [N x n x d] tensor, where 
%                       - N is the number of time series,
%                       - n is the length of the time series
%                       - d is the dimension of the time series
%                     - Only supports equal-length time series.
%                     - For univariate time sries (that is d==1), X is 
%                       simply a [N x n] matrix.
%
%   maxIter         Maximum number of iterations. MM may terminate earlier
%   (o)             default: maxIter = 15
%
%   tol             Tolerance parameter for stopping criterion. The
%   (o)             algorithm stops if the improvement of the Frechet 
%                   variation is below tol.
%                   Default: tol = 1E-4
%
%   initSequence    either empty, a string, a number, or a [n x d] matrix
%   (o)             * default: If initSequence is not defined, a random 
%                     sample of X is selected as initialization.
%                   * if string: must be one of {'random_sample','medoid','euclidean_mean'}
%                      - in such cases a random sample, a medoid or the
%                        Euclidean mean is used as initialization.
%                      - Euclidean mean is only available if n_1==n_2==...==n_N
%                      - note that computing a medoid invoks a quadratic
%                        number of dtw computations
%                   * if number: must be an integer i between 1 and N. 
%                      - in this case the sample X{i} is used as
%                        initialization
%                   * if [n x d] time series, this time series is used as
%                        initialization.
%
%   window          Warping window parameter for dtw distance.
%   (o)
%
% Output
%   mu              Approximate mean time series of size [n x d], 
%                   where n is the length of the initial sequence.
%
%   f               Vector containing Frechet variances f(k) := F(mu^(k))/N,
%   (o)             where mu^(k) is the mean at iteration k.
%                   f(1) is the Frechet variance of the initial sequence
%
% (o) = optional
%__________________
%
% References
% The MM mean algorithm has been introduced in [1]. It is equivalent to the
% DBA algorithm [2].
%
% [1] Schultz, D., & Jain, B. (2018). Nonsmooth analysis and subgradient 
%     methods for averaging in dynamic time warping spaces. 
%     Pattern Recognition, 74, 340-358.
% [2] Petitjean, F., Ketterlin, A., & Ganarski, P. (2011). A global 
%     averaging method for dynamic time warping, with applications to 
%     clustering. Pattern Recognition, 44(3), 678-693.
%
% Author: David Schultz, DAI-Lab, TU Berlin, Germany, 06-2020 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% convert X to cell array if necessary
if ~iscell(X)
   if length(size(X)) > 3
       error('Invalid shape of X. size(X) must be [N,n] or [N,n,d].')
   end
   [N,n,d] = size(X);
   Y = cell(N);
   for i = 1:N
       Y{i} = reshape(X(i,:,:),n,d);
   end
   X=Y;
end

N = length(X);          % number of samples
d = size(X{1},2);       % dimension of data


if nargin < 2 || isempty(maxIter)
    maxIter = 15;
end

if nargin < 3 || isempty(tol)
    tol = 1E-4;
end

if nargin < 4
    initSequence = [];
end

if nargin < 5
    window = [];
end


% initialize mean mu
if isempty(initSequence) || strcmp(initSequence,'random_sample')
    mu = X{randi(N)};
elseif strcmp(initSequence,'medoid')
    mu = medoidSequence(X,window);
elseif strcmp(initSequence,'euclidean_mean')
    mu = euclideanMean(X);
elseif size(initSequence,1) > 1
    mu = initSequence;
elseif length(initSequence) == 1 && initSequence > 0 && initSequence < N+1 && round(initSequence) == initSequence
    mu = X{initSequence};
else
    error('Unknown initialization method')
end
 
n = size(mu,1); % length of the mean

f = zeros(maxIter+1,1);

% Majorize-minimize optimization 
k = 1;
while (k <= 2 || f(k-2)-f(k-1) > tol ) && k <= maxIter
    
% (1) MAJORIZE: find v,w describing an F-majorizing function

    v = zeros(n,1);
    w = zeros(n,d);
    
    frechetVar = 0;
    
    for i = 1:N
        
        x_i = X{i};
        [dist,p] = dtw_(mu,x_i,window);

        [W,V] = getWarpingAndValenceMatrix(p);
        v = v + V;
        w = w + W*x_i;      % componentwise matrix multiplication 
                            % W*x_i = W*x_i(:,j), j=1,...,d
                            
        frechetVar = frechetVar + dist^2/N;
    end
    

    
% (2) MINIMIZE: update the mean
    
    % For local Euclidean distance the mean update can be computed
    % componentwise using the same warping and valence matrix for each 
    % component along the time series dimension j=1,...,d.    
    mu = w./repmat(v,1,d);
    
    f(k) = frechetVar;
    
    k = k+1;
end

if nargout == 2
    f(k) = Frechet(mu,X,window);
    f = f(1:k);
end

end



function x = medoidSequence(X,w) 
%MEDOIDSEQUENCE returns medoid of X
% A medoid is an element of X that minimizes the Frechet function
% among all elements in X
% Input
%   X               Cell array of time series x_i of size [n_i,d], where
%                   n_i is the length of time series x_i and d is the 
%                   uniform dimension of all time series.
%                   example:   X{1} = x_1;  % size [n_1,d]
%                              X{2} = x_2;  % size [n_2,d]
%
%   w               warping window parameter
%
% Output
%   x               Medoid of sample X

N = length(X);
f_min = inf;
i_min = 0;
for i = 1:N
    f = Frechet(X{i},X,w);
    if f < f_min
        f_min = f;
        i_min = i;
    end
end
    x = X{i_min};
end




function x = euclideanMean(X) 
%EUCLIEDEANMEAN returns arithmetic mean of sample X
% Input
%   X               Cell array of time series x_i of size [n_i,d], where
%                   n_i is the length of time series x_i and d is the 
%                   uniform dimension of all time series.
%                   example:   X{1} = x_1;  % size [n_1,d]
%                              X{2} = x_2;  % size [n_2,d]
% 
% Output
%   x               Arithmetic mean of all x_i in X

N = length(X);

x = X{1};
for i = 2:N
    x = x+X{i};
end
x=x/N;

end


function f = Frechet(x,X,w)
%FRECHET computes value of Frechet function
%
% Input
%   x               time series of size [n,d]
% 
%   X               Cell array of time series x_i of size [n_i,d], where
%                   n_i is the length of time series x_i and d is the 
%                   uniform dimension of all time series.
%                   example:   X{1} = x_1;  % size [n_1,d]
%                              X{2} = x_2;  % size [n_2,d]
%
%   w               warping window parameter
%
% Output
%    f              Value of Frechet function at x wrt. X.
%                   f = F(x), where
%                                   N
%                       F(x) = 1/N sum dtw^2(x,x_i).
%                                  i=1
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

N=length(X);
f = 0;
for i = 1:N
    d = dtw_(x,X{i},w);
    f = f + d^2;
end
f=f/N;
end


